==> release-cli-v0.16.0.tar.bz2: release-cli-v0.16.0/internal/commands/create_test.go <==
package commands

import (
	"flag"
	"io"
	"os"
	"testing"

	"github.com/sirupsen/logrus"
	testlog "github.com/sirupsen/logrus/hooks/test"
	"github.com/stretchr/testify/mock"
	"github.com/stretchr/testify/require"
	"github.com/urfave/cli/v2"

	"gitlab.com/gitlab-org/release-cli/internal/flags"
	"gitlab.com/gitlab-org/release-cli/internal/gitlab"
	"gitlab.com/gitlab-org/release-cli/internal/testdata"
	"gitlab.com/gitlab-org/release-cli/internal/testhelpers"
)

func TestCreate(t *testing.T) {
	logger, hook := testlog.NewNullLogger()

	mhc := &gitlab.MockHTTPClient{}
	mhc.On("Do", mock.Anything).Return(testdata.Responses[testdata.ResponseCreateReleaseSuccess](), nil).Once()

	defer mhc.AssertExpectations(t)

	create, ctx := newTestCreate(t, logrus.NewEntry(logger), mhc)
	require.Equal(t, "create", create.Name)
	require.Len(t, create.Flags, 8)

	err := create.Run(ctx)
	require.NoError(t, err)

	for _, entry := range hook.AllEntries() {
		require.Contains(t, []string{"Creating Release...", "file does not exist, using string value for --description", "release created successfully!"}, entry.Message)
	}
}

func TestCreateFromFile(t *testing.T) {
	logger, hook := testlog.NewNullLogger()

	mhc := &gitlab.MockHTTPClient{}
	mhc.On("Do", mock.Anything).Return(testdata.Responses[testdata.ResponseCreateReleaseSuccess](), nil).Once()

	defer mhc.AssertExpectations(t)

	createFromFile, ctx := newTestCreateFromFile(t, logrus.NewEntry(logger), mhc)
	require.Equal(t, "create-from-file", createFromFile.Name)
	require.Len(t, createFromFile.Flags, 9)

	err := createFromFile.Run(ctx)
	require.NoError(t, err)

	for _, entry := range hook.AllEntries() {
		require.Contains(t, []string{"Creating Release...", "file does not exist, using string value for --description", "release created successfully!"}, entry.Message)
	}
}

func TestCreateAssets(t *testing.T) {
	logger, _ := testlog.NewNullLogger()

	tests := []struct {
		name           string
		extraArgs      []string
		expectedRunErr bool
		expectedErrMsg string
	}{
		{
			name:      "success_with_assets_link_json",
			extraArgs: []string{"--assets-link", `{"name":"Asset1","url":"https://<domain>/some/location/1","link_type":"other","filepath":"xzy1"}`},
		},
		{
			name: "success_with_multiple_assets_link_json",
			extraArgs: []string{"--assets-link", `{"name":"Asset1","url":"https://<domain>/some/location/1","link_type":"other","filepath":"xzy1"}`,
				"--assets-link", `{"name":"Asset2","url":"https://<domain>/some/location/2","link_type":"other","filepath":"xzy2"}`,
				"--assets-link", `{"name":"Asset3","url":"https://<domain>/some/location/3","link_type":"other","filepath":"xzy3"}`},
		},
		{
			name:           "failed_with_invalid_assets_link_json",
			extraArgs:      []string{"--assets-link", "not_json"},
			expectedErrMsg: "new CreateReleaseRequest: failed to parse assets: 1 error occurred:\n\t* invalid delimiter for asset: \"not_json\"\n\n",
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			mhc := gitlab.MockHTTPClient{}
			if tt.expectedErrMsg == "" {
				mhc.On("Do", mock.Anything).Return(testdata.Responses[testdata.ResponseCreateReleaseSuccess](), nil).Once()
				defer mhc.AssertExpectations(t)
			}

			create, ctx := newTestCreate(t, logrus.NewEntry(logger), &mhc, tt.extraArgs...)
			err := create.Run(ctx)
			if tt.expectedErrMsg != "" {
				require.EqualError(t, err, tt.expectedErrMsg)
			} else {
				require.NoError(t, err)
			}
		})
	}
}

func TestCreateMilestonesAndReleasedAt(t *testing.T) {
	logger, _ := testlog.NewNullLogger()

	tests := []struct {
		name           string
		extraArgs      []string
		expectedRunErr bool
		expectedErrMsg string
	}{
		{
			name:      "success_milestone_and_released_at",
			extraArgs: []string{"--milestone", "v1.0", "--released-at", "2019-01-03T01:55:18.203Z"},
		},
		{
			name:      "success_multiple_milestones",
			extraArgs: []string{"--milestone", "v1.0", "--milestone", "v1.0-rc"},
		},
		{
			name:      "success_released_at",
			extraArgs: []string{"--released-at", "2019-01-03T01:55:18.203Z"},
		},
		{
			name:           "invalid_released_at",
			extraArgs:      []string{"--released-at", "2019/01/03 01:55:18"},
			expectedErrMsg: "failed to parse released-at: parsing time",
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			mhc := gitlab.MockHTTPClient{}
			defer mhc.AssertExpectations(t)
			if tt.expectedErrMsg == "" {
				mhc.On("Do", mock.Anything).Return(testdata.Responses[testdata.ResponseCreateReleaseSuccess](), nil).Once()
			}

			create, ctx := newTestCreate(t, logrus.NewEntry(logger), &mhc, tt.extraArgs...)
			err := create.Run(ctx)
			if tt.expectedErrMsg != "" {
				require.Error(t, err)
				require.Contains(t, err.Error(), tt.expectedErrMsg)
				return
			}

			require.NoError(t, err)
		})
	}
}

func TestGetDescription(t *testing.T) {
	chdirSet := false
	chdir := testhelpers.ChdirInPath(t, "../", &chdirSet)

	defer chdir()

	logger, _ := testlog.NewNullLogger()

	file := "testdata/description.txt"
	content, err := os.ReadFile(file)
	require.NoError(t, err)

	tests := []struct {
		name     string
		arg      string
		expected string
	}{
		{
			name:     "description_from_string_without_whitespace",
			arg:      "description",
			expected: "description",
		},
		{
			name:     "description_from_string_with_whitespace",
			arg:      "description from string",
			expected: "description from string",
		},
		{
			name:     "description_from_empty_string",
			arg:      "",
			expected: "",
		},
		{
			name:     "description_from_file",
			arg:      file,
			expected: string(content),
		},
		{
			name:     "description_from_file_with_path_traversal_use_as_string",
			arg:      "../testdata/description-traversal.txt",
			expected: "../testdata/description-traversal.txt",
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			description, err := getDescription(tt.arg, logger)

			require.NoError(t, err)
			require.Equal(t, tt.expected, description)
		})
	}
}

func Test_newCreateReleaseReq(t *testing.T) {
	logger, _ := testlog.NewNullLogger()

	rat, err := gitlab.ParseDateTime("2019-01-03T01:55:18.203Z")
	require.NoError(t, err)

	baseCRR := gitlab.CreateReleaseRequest{
		ID:          "",
		Name:        "name",
		Description: "desc",
		TagName:     "tag",
		TagMessage:  "tag message",
		Ref:         "ref",
		Assets: &gitlab.Assets{Links: []*gitlab.Link{
			{Name: "asset-1", URL: "asset-url-1"},
		}},
		Milestones: []string{"m1", "m2"},
		ReleasedAt: &rat,
	}

	requiredArgs := []string{"--name", "name", "--description", "desc", "--tag-name", "tag", "--ref", "ref"}

	tests := []struct {
		name           string
		args           []string
		expected       *gitlab.CreateReleaseRequest
		expectedErrMsg string
	}{
		{
			name: "all_valid_fields",
			args: []string{
				"--assets-link", `{"name": "asset-1", "url":"asset-url-1" }`,
				"--milestone", "m1", "--milestone", "m2", "--released-at", "2019-01-03T01:55:18.203Z",
				"--tag-message", "tag message",
			},
			expected: &baseCRR,
		},
		{
			name: "no_assets",
			args: []string{
				"--milestone", "m1", "--milestone", "m2", "--released-at", "2019-01-03T01:55:18.203Z",
				"--tag-message", "tag message",
			},
			expected: newTestCreateReleaseRequest(baseCRR, func(crr *gitlab.CreateReleaseRequest) {
				crr.Assets = nil
			}),
		},
		{
			name: "no_released_at",
			args: []string{
				"--assets-link", `{"name": "asset-1", "url":"asset-url-1" }`,
				"--milestone", "m1", "--milestone", "m2",
				"--tag-message", "tag message",
			},
			expected: newTestCreateReleaseRequest(baseCRR, func(crr *gitlab.CreateReleaseRequest) {
				crr.ReleasedAt = nil
			}),
		},
		{
			name: "no_milestones",
			args: []string{
				"--assets-link", `{"name": "asset-1", "url":"asset-url-1" }`,
				"--released-at", "2019-01-03T01:55:18.203Z",
				"--tag-message", "tag message",
			},
			expected: newTestCreateReleaseRequest(baseCRR, func(crr *gitlab.CreateReleaseRequest) {
				crr.Milestones = make([]string, 0)
			}),
		},
		{
			name: "no_extra_args",
			args: []string{},
			expected: newTestCreateReleaseRequest(baseCRR, func(crr *gitlab.CreateReleaseRequest) {
				crr.Assets = nil
				crr.Milestones = make([]string, 0)
				crr.ReleasedAt = nil
				crr.TagMessage = ""
			}),
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			args := append(requiredArgs, tt.args...)
			set := getFlagSet(t, args...)
			ctx := cli.NewContext(nil, set, nil)

			got, err := newCreateReleaseReq(ctx, logger)
			if tt.expectedErrMsg != "" {
				require.Error(t, err)
				require.Contains(t, err.Error(), tt.expectedErrMsg)
				return
			}
			require.NoError(t, err)

			require.Equal(t, tt.expected, got)
		})
	}
}

// to mock cli.Context we need to manually create the flagSet
func getFlagSet(t *testing.T, args ...string) *flag.FlagSet {
	t.Helper()

	set := flag.NewFlagSet(t.Name(), 0)
	al := cli.StringSliceFlag{Name: flags.AssetsLink}
	_ = al.Apply(set)
	m := cli.StringSliceFlag{Name: flags.Milestone}
	_ = m.Apply(set)

	set.String(flags.Name, "", "")
	set.String(flags.Description, "", "")
	set.String(flags.TagName, "", "")
	set.String(flags.TagMessage, "", "")
	set.String(flags.Ref, "", "")
	set.String(flags.ReleasedAt, "", "")

	err := set.Parse(args)
	require.NoError(t, err)

	return set
}

func newTestCreateReleaseRequest(base gitlab.CreateReleaseRequest, mutators ...func(*gitlab.CreateReleaseRequest)) *gitlab.CreateReleaseRequest {
	for _, mutate := range mutators {
		mutate(&base)
	}

	return &base
}

func newTestCreate(t *testing.T, logger logrus.FieldLogger, mhc gitlab.HTTPClient, extraArgs ...string) (*cli.Command, *cli.Context) {
	t.Helper()

	clientFn := func(ctx *cli.Context, log logrus.FieldLogger) (gitlab.HTTPClient, error) {
		return mhc, nil
	}

	args := []string{"create", "--name", t.Name(), "--description", t.Name(), "--tag-name", t.Name()}
	args = append(args, extraArgs...)

	app := &cli.App{Writer: io.Discard}
	set := flag.NewFlagSet(t.Name(), 0)
	set.String(flags.ServerURL, "https://gitlab.com", "")
	set.String(flags.JobToken, "token", "")

	err := set.Parse(args)
	require.NoError(t, err)

	create := Create(logger, clientFn)
	ctx := cli.NewContext(app, set, nil)

	return create, ctx
}

func newTestCreateFromFile(t *testing.T, logger logrus.FieldLogger, mhc gitlab.HTTPClient, extraArgs ...string) (*cli.Command, *cli.Context) {
	t.Helper()

	clientFn := func(ctx *cli.Context, log logrus.FieldLogger) (gitlab.HTTPClient, error) {
		return mhc, nil
	}

	args := []string{"create-from-file", "--file", "../testdata/release.yml"}
	args = append(args, extraArgs...)

	app := &cli.App{Writer: io.Discard}
	set := flag.NewFlagSet(t.Name(), 0)
	set.String(flags.ServerURL, "https://gitlab.com", "")
	set.String(flags.JobToken, "token", "")

	err := set.Parse(args)
	require.NoError(t, err)

	create := CreateFromFile(logger, clientFn)
	ctx := cli.NewContext(app, set, nil)

	return create, ctx
}

==> release-cli-v0.19.0.tar.bz2: release-cli-v0.19.0/internal/commands/create_test.go <==
package commands

import (
	"encoding/json"
	"flag"
	"io"
	http "net/http"
	"os"
	"testing"
	"time"

	"github.com/sirupsen/logrus"
	testlog "github.com/sirupsen/logrus/hooks/test"
	"github.com/stretchr/testify/mock"
	"github.com/stretchr/testify/require"
	"github.com/urfave/cli/v2"

	"gitlab.com/gitlab-org/release-cli/internal/flags"
	"gitlab.com/gitlab-org/release-cli/internal/gitlab"
	"gitlab.com/gitlab-org/release-cli/internal/testdata"
	"gitlab.com/gitlab-org/release-cli/internal/testhelpers"
)

func TestCreate(t *testing.T) {
	logger, hook := testlog.NewNullLogger()

	mhc := &gitlab.MockHTTPClient{}
	mhc.On("Do", mock.Anything).Return(testdata.Responses[testdata.ResponseCreateReleaseSuccess](), nil).Once()

	defer mhc.AssertExpectations(t)

	create, ctx := newTestCreate(t, logrus.NewEntry(logger), mhc)
	require.Equal(t, "create", create.Name)
	require.Len(t, create.Flags, 9)

	err := create.Run(ctx)
	require.NoError(t, err)

	for _, entry := range hook.AllEntries() {
		require.Contains(t, []string{"Creating Release...", "file does not exist, using string value for --description", "release created successfully!"}, entry.Message)
	}
}

func TestCreateFromFile(t *testing.T) {
	releasedAt, _ := time.Parse(time.RFC3339, "2019-01-03T01:55:18.203Z")
	expectedRelease := gitlab.CreateReleaseRequest{
		Name:        "Release for test",
		Description: "This is a release for testing only.",
		TagName:     getEnvOrDefault("CI_COMMIT_TAG", "v1.2.3"),
		ReleasedAt:  &releasedAt,
		Ref:         getEnvOrDefault("CI_COMMIT_SHA", "abcdefgh12345678"),
		Assets: &gitlab.Assets{
			Links: []*gitlab.Link{
				{
					Name:            "asset1",
					URL:             "https://gitlab.com/gitlab-org/release-cli",
					DirectAssetPath: "/bin/binary_file",
				},
				{
					Name: "asset2",
					URL:  "https://gitlab.com/",
				},
			},
		},
		Milestones: []string{"v2.0", "v3.0"},
	}

	logger, hook := testlog.NewNullLogger()

	mhc := &gitlab.MockHTTPClient{}
	mhc.On("Do", mock.Anything).Return(testdata.Responses[testdata.ResponseCreateReleaseSuccess](), nil).Once()

	defer mhc.AssertExpectations(t)

	createFromFile, ctx := newTestCreateFromFile(t, logrus.NewEntry(logger), mhc)
	require.Equal(t, "create-from-file", createFromFile.Name)
	require.Len(t, createFromFile.Flags, 10)

	err := createFromFile.Run(ctx)
	require.NoError(t, err)

	// test the release passed to the HTTPClient since the response is mocked
	release := getReleaseFromRequest(t, mhc)
	require.Equal(t, expectedRelease, release)

	for _, entry := range hook.AllEntries() {
		require.Contains(t, []string{"Creating Release...", "release created successfully!"}, entry.Message)
	}
}

func TestCreateAssets(t *testing.T) {
	logger, _ := testlog.NewNullLogger()

	tests := []struct {
		name           string
		extraArgs      []string
		expectedRunErr bool
		expectedErrMsg string
	}{
		{
			name:      "success_with_assets_link_json",
			extraArgs: []string{"--assets-link", `{"name":"Asset1","url":"https://<domain>/some/location/1","link_type":"other","direct_asset_path":"xzy1"}`},
		},
		{
			name: "success_with_multiple_assets_link_json",
			extraArgs: []string{"--assets-link", `{"name":"Asset1","url":"https://<domain>/some/location/1","link_type":"other","direct_asset_path":"xzy1"}`,
				"--assets-link", `{"name":"Asset2","url":"https://<domain>/some/location/2","link_type":"other","direct_asset_path":"xzy2"}`,
				"--assets-link", `{"name":"Asset3","url":"https://<domain>/some/location/3","link_type":"other","direct_asset_path":"xzy3"}`},
		},
		{
			name:           "failed_with_invalid_assets_link_json",
			extraArgs:      []string{"--assets-link", "not_json"},
			expectedErrMsg: "new CreateReleaseRequest: failed to parse assets: 1 error occurred:\n\t* invalid delimiter for asset: \"not_json\"\n\n",
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			mhc := gitlab.MockHTTPClient{}
			if tt.expectedErrMsg == "" {
				mhc.On("Do", mock.Anything).Return(testdata.Responses[testdata.ResponseCreateReleaseSuccess](), nil).Once()
				defer mhc.AssertExpectations(t)
			}

			create, ctx := newTestCreate(t, logrus.NewEntry(logger), &mhc, tt.extraArgs...)
			err := create.Run(ctx)
			if tt.expectedErrMsg != "" {
				require.EqualError(t, err, tt.expectedErrMsg)
			} else {
				require.NoError(t, err)
			}
		})
	}
}

func TestCreateMilestonesAndReleasedAt(t *testing.T) {
	logger, _ := testlog.NewNullLogger()

	tests := []struct {
		name           string
		extraArgs      []string
		expectedRunErr bool
		expectedErrMsg string
	}{
		{
			name:      "success_milestone_and_released_at",
			extraArgs: []string{"--milestone", "v1.0", "--released-at", "2019-01-03T01:55:18.203Z"},
		},
		{
			name:      "success_multiple_milestones",
			extraArgs: []string{"--milestone", "v1.0", "--milestone", "v1.0-rc"},
		},
		{
			name:      "success_released_at",
			extraArgs: []string{"--released-at", "2019-01-03T01:55:18.203Z"},
		},
		{
			name:           "invalid_released_at",
			extraArgs:      []string{"--released-at", "2019/01/03 01:55:18"},
			expectedErrMsg: "failed to parse released-at: parsing time",
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			mhc := gitlab.MockHTTPClient{}
			defer mhc.AssertExpectations(t)
			if tt.expectedErrMsg == "" {
				mhc.On("Do", mock.Anything).Return(testdata.Responses[testdata.ResponseCreateReleaseSuccess](), nil).Once()
			}

			create, ctx := newTestCreate(t, logrus.NewEntry(logger), &mhc, tt.extraArgs...)
			err := create.Run(ctx)
			if tt.expectedErrMsg != "" {
				require.Error(t, err)
				require.Contains(t, err.Error(), tt.expectedErrMsg)
				return
			}

			require.NoError(t, err)
		})
	}
}

func TestCreateCatalogPublish(t *testing.T) {
	logger, _ := testlog.NewNullLogger()

	tests := []struct {
		name      string
		extraArgs []string
	}{
		{
			name:      "catalog_publish_true",
			extraArgs: []string{"--catalog-publish"},
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			mhc := gitlab.MockHTTPClient{}
			mhc.On("Do", mock.Anything).Return(testdata.Responses[testdata.ResponseCreateReleaseSuccess](), nil).Once()
			defer mhc.AssertExpectations(t)

			create, ctx := newTestCreate(t, logrus.NewEntry(logger), &mhc, tt.extraArgs...)
			err := create.Run(ctx)
			require.NoError(t, err)
		})
	}
}

func TestGetDescription(t *testing.T) {
	chdirSet := false
	chdir := testhelpers.ChdirInPath(t, "../", &chdirSet)

	defer chdir()

	logger, _ := testlog.NewNullLogger()

	file := "testdata/description.txt"
	content, err := os.ReadFile(file)
	require.NoError(t, err)

	tests := []struct {
		name     string
		arg      string
		expected string
	}{
		{
			name:     "description_from_string_without_whitespace",
			arg:      "description",
			expected: "description",
		},
		{
			name:     "description_from_string_with_whitespace",
			arg:      "description from string",
			expected: "description from string",
		},
		{
			name:     "description_from_empty_string",
			arg:      "",
			expected: "",
		},
		{
			name:     "description_from_file",
			arg:      file,
			expected: string(content),
		},
		{
			name:     "description_from_file_with_path_traversal_use_as_string",
			arg:      "../testdata/description-traversal.txt",
			expected: "../testdata/description-traversal.txt",
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			description, err := getDescription(tt.arg, logger)

			require.NoError(t, err)
			require.Equal(t, tt.expected, description)
		})
	}
}

func Test_newCreateReleaseReq(t *testing.T) {
	logger, _ := testlog.NewNullLogger()

	rat, err := gitlab.ParseDateTime("2019-01-03T01:55:18.203Z")
	require.NoError(t, err)

	trueVal := true

	baseCRR := gitlab.CreateReleaseRequest{
		ID:          "",
		Name:        "name",
		Description: "desc",
		TagName:     "tag",
		TagMessage:  "tag message",
		Ref:         "ref",
		Assets: &gitlab.Assets{Links: []*gitlab.Link{
			{Name: "asset-1", URL: "asset-url-1"},
		}},
		Milestones: []string{"m1", "m2"},
		ReleasedAt: &rat,
	}

	requiredArgs := []string{"--name", "name", "--description", "desc", "--tag-name", "tag", "--ref", "ref"}

	tests := []struct {
		name           string
		args           []string
		expected       *gitlab.CreateReleaseRequest
		expectedErrMsg string
	}{
		{
			name: "all_valid_fields",
			args: []string{
				"--assets-link", `{"name": "asset-1", "url":"asset-url-1" }`,
				"--milestone", "m1", "--milestone", "m2", "--released-at", "2019-01-03T01:55:18.203Z",
				"--tag-message", "tag message",
			},
			expected: &baseCRR,
		},
		{
			name: "no_assets",
			args: []string{
				"--milestone", "m1", "--milestone", "m2", "--released-at", "2019-01-03T01:55:18.203Z",
				"--tag-message", "tag message",
			},
			expected: newTestCreateReleaseRequest(baseCRR, func(crr *gitlab.CreateReleaseRequest) {
				crr.Assets = nil
			}),
		},
		{
			name: "no_released_at",
			args: []string{
				"--assets-link", `{"name": "asset-1", "url":"asset-url-1" }`,
				"--milestone", "m1", "--milestone", "m2",
				"--tag-message", "tag message",
			},
			expected: newTestCreateReleaseRequest(baseCRR, func(crr *gitlab.CreateReleaseRequest) {
				crr.ReleasedAt = nil
			}),
		},
		{
			name: "no_milestones",
			args: []string{
				"--assets-link", `{"name": "asset-1", "url":"asset-url-1" }`,
				"--released-at", "2019-01-03T01:55:18.203Z",
				"--tag-message", "tag message",
			},
			expected: newTestCreateReleaseRequest(baseCRR, func(crr *gitlab.CreateReleaseRequest) {
				crr.Milestones = make([]string, 0)
			}),
		},
		{
			name: "no_extra_args",
			args: []string{},
			expected: newTestCreateReleaseRequest(baseCRR, func(crr *gitlab.CreateReleaseRequest) {
				crr.Assets = nil
				crr.Milestones = make([]string, 0)
				crr.ReleasedAt = nil
				crr.TagMessage = ""
			}),
		},
		{
			name: "with_catalog_publish",
			args: []string{
				"--catalog-publish",
			},
			expected: newTestCreateReleaseRequest(baseCRR, func(crr *gitlab.CreateReleaseRequest) {
				crr.LegacyCatalogPublish = &trueVal
				crr.Assets = nil
				crr.Milestones = make([]string, 0)
				crr.ReleasedAt = nil
				crr.TagMessage = ""
			}),
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			args := append(requiredArgs, tt.args...)
			set := getFlagSet(t, args...)
			ctx := cli.NewContext(nil, set, nil)

			got, err := newCreateReleaseReq(ctx, logger)
			if tt.expectedErrMsg != "" {
				require.Error(t, err)
				require.Contains(t, err.Error(), tt.expectedErrMsg)
				return
			}
			require.NoError(t, err)

			require.Equal(t, tt.expected, got)
		})
	}
}

// to mock cli.Context we need to manually create the flagSet
func getFlagSet(t *testing.T, args ...string) *flag.FlagSet {
	t.Helper()

	set := flag.NewFlagSet(t.Name(), 0)
	al := cli.StringSliceFlag{Name: flags.AssetsLink}
	_ = al.Apply(set)
	m := cli.StringSliceFlag{Name: flags.Milestone}
	_ = m.Apply(set)

	set.String(flags.Name, "", "")
	set.String(flags.Description, "", "")
	set.String(flags.TagName, "", "")
	set.String(flags.TagMessage, "", "")
	set.String(flags.Ref, "", "")
	set.String(flags.ReleasedAt, "", "")
	set.Bool(flags.CatalogPublish, false, "")

	err := set.Parse(args)
	require.NoError(t, err)

	return set
}

func newTestCreateReleaseRequest(base gitlab.CreateReleaseRequest, mutators ...func(*gitlab.CreateReleaseRequest)) *gitlab.CreateReleaseRequest {
	for _, mutate := range mutators {
		mutate(&base)
	}

	return &base
}

func newTestCreate(t *testing.T, logger logrus.FieldLogger, mhc gitlab.HTTPClient, extraArgs ...string) (*cli.Command, *cli.Context) {
	t.Helper()

	clientFn := func(ctx *cli.Context, log logrus.FieldLogger) (gitlab.HTTPClient, error) {
		return mhc, nil
	}

	args := []string{"create", "--name", t.Name(), "--description", t.Name(), "--tag-name", t.Name()}
	args = append(args, extraArgs...)

	app := &cli.App{Writer: io.Discard}
	set := flag.NewFlagSet(t.Name(), 0)
	set.String(flags.ServerURL, "https://gitlab.com", "")
	set.String(flags.JobToken, "token", "")

	err := set.Parse(args)
	require.NoError(t, err)

	create := Create(logger, clientFn)
	ctx := cli.NewContext(app, set, nil)

	return create, ctx
}

func newTestCreateFromFile(t *testing.T, logger logrus.FieldLogger, mhc gitlab.HTTPClient, extraArgs ...string) (*cli.Command, *cli.Context) {
	t.Helper()

	clientFn := func(ctx *cli.Context, log logrus.FieldLogger) (gitlab.HTTPClient, error) {
		return mhc, nil
	}

	args := []string{"create-from-file", "--file", "../testdata/release.yml"}
	args = append(args, extraArgs...)

	app := &cli.App{Writer: io.Discard}
	set := flag.NewFlagSet(t.Name(), 0)
	set.String(flags.ServerURL, "https://gitlab.com", "")
	set.String(flags.JobToken, "token", "")

	err := set.Parse(args)
	require.NoError(t, err)

	create := CreateFromFile(logger, clientFn)
	ctx := cli.NewContext(app, set, nil)

	return create, ctx
}

// getReleaseFromRequest reads the request body from the mock server call and
// unmarshals it into a CreateReleaseRequest
func getReleaseFromRequest(t *testing.T, mhc *gitlab.MockHTTPClient) gitlab.CreateReleaseRequest {
	// check that mock was called once with an http request
	require.Len(t, mhc.Mock.Calls, 1)
	require.Len(t, mhc.Mock.Calls[0].Arguments, 1)
	request := mhc.Mock.Calls[0].Arguments[0].(*http.Request)
	require.NotNil(t, request)

	body, err := request.GetBody()
	require.NoError(t, err)

	data, err := io.ReadAll(body)
	require.NoError(t, err)

	release := gitlab.CreateReleaseRequest{}
	err = json.Unmarshal(data, &release)
	require.NoError(t, err)

	return release
}

// getEnvOrDefault returns the value of an environment variable if it exists,
// otherwise the provided default value
func getEnvOrDefault(env string, val string) string {
	result := os.Getenv(env)
	if result == "" {
		result = val
	}

	return result
}
